###############################################################################
#
# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
# Reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

__revision__ = "$Id: reporting.py 14 2008-07-14 22:09:55Z nicoe $"
__metaclass__ = type

import os, sys
import warnings
import cStringIO
from genshi.template import MarkupTemplate, TextTemplate
from genshi.template import TemplateLoader

from templates import NullTemplate
from templates.odt import Template as OOTemplate

try:
    from templates.pdf import Template as PDFTemplate
except ImportError:
    PDFTemplate = NullTemplate
    warnings.warn("trml2pdf is not installed on your system. You will not "
                  "be able to create PDF files.")

def _absolute(path):
    "Compute the absolute path of path relative to the caller file"
    if os.path.isabs(path):
        return path
    caller_fname = sys._getframe(2).f_globals['__file__']
    dir = os.path.dirname(caller_fname)
    return os.path.abspath(os.path.join(dir, path))

def _guess_type(mime):
    mime = mime.lower()
    type, stype = mime.split('/', 1)
    if type == 'application':
        if 'opendocument' in stype:
            return 'oo.org'
        else:
            return stype
    elif type == 'text':
        if stype in ('xml', 'html', 'xhtml'):
            return 'markup'
        else:
            return 'text'


class MIMETemplateLoader(TemplateLoader):
    """This subclass of TemplateLoader use mimetypes to search and find
    templates to load.
    """

    factories = {'pdf': PDFTemplate,
                 'oo.org': OOTemplate,
                 'text': TextTemplate,
                 'markup': MarkupTemplate}

    mime_func = [_guess_type]

    def get_type(self, mime):
        for func in reversed(self.mime_func):
            t = func(mime)
            if t is not None:
                return t

    def load(self, path, mime):
        rtype = self.get_type(mime)
        return super(MIMETemplateLoader, self).load(path,
                                                    cls=self.factories[rtype])

    @classmethod
    def add_factory(cls, abbr_mimetype, template_factory, id_function=None):
        """adds a template factory to the already known factories"""
        if abbr_mimetype in cls.factories:
            warnings.warn('You are overriding an already defined link.')
        cls.factories[abbr_mimetype] = template_factory
        if id_function is not None:
            cls.mime_func.append(id_function)


class Report:
    """Report is a simple interface on top of a rendering template.
    """

    def __init__(self, path, mimetype, factory, loader):
        self.fpath = path
        self.mimetype = mimetype
        self.data_factory = factory
        self.tmpl_loader = loader

    def __call__(self, obj, **kwargs):
        template = self.tmpl_loader.load(self.fpath, self.mimetype)
        data = self.data_factory(obj, **kwargs)
        report = template.generate(**data)
        if not isinstance(report, (cStringIO.InputType, cStringIO.OutputType)):
            report = StringIO(str(report))
        return report

    def __repr__(self):
        return '<relatorio report on %s>' % self.fpath


class DefaultFactory:

    def __call__(self, obj, **kwargs):
        data = {}
        data['o'] = obj
        data['args'] = kwargs
        return data


class ReportRepository:
    """ReportRepository stores the report definition associated to objects.

    The report are indexed in this object by the object class they are working
    on and the name given to it by the user.
    """

    def __init__(self, datafactory=DefaultFactory):
        self.reports = {}
        self.default_factory = datafactory
        self.loader = MIMETemplateLoader(auto_reload=True)

    def add_report(self, klass, mimetype, template_path, data_factory=None,
                   report_name='default'):
        if data_factory is None:
            data_factory = self.default_factory
        reports = self.reports.setdefault(klass, {})
        report = Report(_absolute(template_path), mimetype, data_factory(),
                        self.loader)
        reports[report_name] = mimetype, report
        reports.setdefault(mimetype, []).append((report_name, report))




###############################################################################
#
# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
# Reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

__revision__ = "$Id: test_api.py 1 2008-07-04 14:31:52Z nicoe $"

import os
from nose.tools import *

from reporting import (ReportRepository, Report, MIMETemplateLoader,
                       DefaultFactory, _absolute)

class StubObject(object):

    def __init__(self, **kwargs):
        for key, val in kwargs.iteritems():
            setattr(self, key, val)


class TestRepository(object):

    def test_register(self):
        "Testing the registration"
        reporting = ReportRepository()
        reporting.add_report(StubObject, 'text/plain', 
                             os.path.join('templates', 'test.tmpl'))

        assert_true(StubObject in reporting.reports)
        assert_true('default' in reporting.reports[StubObject])
        assert_true('text/plain' in reporting.reports[StubObject])

        mime, report = reporting.reports[StubObject]['default']
        eq_(mime, 'text/plain')
        eq_(report.mimetype, 'text/plain')
        assert_true(report.fpath.endswith(os.path.join('templates', 
                                                       'test.tmpl')))

        name, report2 = reporting.reports[StubObject]['text/plain'][0]
        eq_(name, 'default')
        eq_(report, report2)

    def test_absolute(self):
        "Test the absolute path calculation"
        our_dir, _ = os.path.split(__file__)
        new_path = _absolute(os.path.join('brol', 'toto'))
        eq_(os.path.join(our_dir, 'brol', 'toto'), new_path)


class TestReport(object):

    def setup(self):
        self.loader = MIMETemplateLoader()
        our_dir, _ = os.path.split(__file__)
        self.report = Report(os.path.join(our_dir, 'templates', 'test.tmpl'),
                             'text/plain', DefaultFactory(), self.loader)

    def test_report(self):
        "Testing the report generation"
        a = StubObject(name='OpenHex')
        eq_(self.report(a).getvalue(), 'Hello OpenHex.\n')

    def test_factory(self):
        "Testing the data factory"
        class MyFactory:
            def __call__(self, obj, time, y=1):
                d = dict()
                d['o'] = obj
                d['y'] = y
                d['time'] = time
                d['func'] = lambda x: x+1
                return d

        our_dir, _ = os.path.split(__file__)
        report = Report(os.path.join(our_dir, 'templates', 'time.tmpl'),
                        'text/plain', MyFactory(), self.loader)

        a = StubObject(name='Foo')
        eq_(report(a, time="One o'clock").getvalue(), 
            "Hi Foo,\nIt's One o'clock to 2 !\n")
        eq_(report(a, time="One o'clock", y=4).getvalue(), 
            "Hi Foo,\nIt's One o'clock to 5 !\n")
        assert_raises(TypeError, report, a) 




###############################################################################
#
# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
# Reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

__revision__ = "$Id: odt.py 14 2008-07-14 22:09:55Z nicoe $"
__metaclass__ = type

import os
import re
import md5
import zipfile
from cStringIO import StringIO

import lxml.etree
from genshi.template import Template as GenshiTemplate, MarkupTemplate

NS = {'text': 'urn:oasis:names:tc:opendocument:xmlns:text:1.0',
      'table': 'urn:oasis:names:tc:opendocument:xmlns:table:1.0',
      'draw': 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0',
      'xlink': 'http://www.w3.org/1999/xlink',
      'py': 'http://genshi.edgewall.org/',
      }
GENSHI_TAGS = re.compile(r'<(/)?(for|choose|otherwise|when|if|with)( (\w+)="(.*)"|)>')
EXTENSIONS = {'image/png': 'png',
              'image/jpeg': 'jpg',
              'image/bmp': 'bmp',
              'image/gif': 'gif',
              'image/tiff': 'tif',
              'image/xbm': 'xbm',
             }

class ImageHref:
    
    def __init__(self, zipfile):
        self.zip = zipfile
        self.count = 0

    def __call__(self, expr, name):
        bitstream, mimetype = expr
        name = md5.new('%s-%s' % (name, self.count)).hexdigest()
        self.count += 1
        path = 'Pictures/%s.%s' % (name, EXTENSIONS[mimetype])
        bitstream.seek(0)
        self.zip.writestr(path, bitstream.read())
        return {'{%s}href' % NS['xlink']: path}


class Template(GenshiTemplate):

    def __init__(self, source, filepath=None, filename=None, loader=None,
                 encoding=None, lookup='strict', allow_exec=True):
        inzip = zipfile.ZipFile(filepath)
        content = inzip.read('content.xml')
        inzip.close()
        self.content_template = MarkupTemplate(self.add_directives(content))
        super(Template, self).__init__(source, filepath, filename, loader,
                                       encoding, lookup, allow_exec)

    def _parse(self, source, encoding):
        pass

    def _prepare(self, stream):
        return []

    def add_directives(self, content):
        tree = lxml.etree.parse(StringIO(content))
        root = tree.getroot()

        self._handle_placeholders(tree)
        self._handle_images(tree)
        return lxml.etree.tostring(tree)

    def _handle_placeholders(self, tree):
        """
        Will treat all placeholders tag (py:if/for/choose/when/otherwise)
        tags
        """
        # First we create the list of all the placeholders nodes.
        # If this is node matches a genshi directive it is put apart for
        # further processing.
        genshi_directives, placeholders = [], []
        for statement in tree.xpath('//text:placeholder', namespaces=NS):
            match_obj = GENSHI_TAGS.match(statement.text)
            if match_obj is not None:
                genshi_directives.append(statement)
            placeholders.append((statement, match_obj))

        # Then we match the opening and closing directives together
        idx = 0
        genshi_pairs, inserted = [], []
        for statement in genshi_directives:
            if not statement.text.startswith('</'):
                genshi_pairs.append([statement, None])
                inserted.append(idx)
                idx += 1
            else:
                genshi_pairs[inserted.pop()][1] = statement

        for p, match_obj in placeholders:
            if match_obj is not None:
                c_dir, directive, _, attr, a_val = match_obj.groups()
            else:
                c_dir, directive, _, attr, a_val = (None, p.text[1:-1], None,
                                                    None, None)

            if p in genshi_directives:
                # If the placeholder is a for statement:
                #    - we operate only on opening statement
                #    - we find the nearest ancestor of the closing and opening
                #      statement
                #    - we create a <py:for> node
                #    - we add all the node between the opening and closing
                #      statements to this new node
                #    - we replace the opening statement by the <py:for> node
                #    - we delete the closing statement

                if c_dir is not None:
                    continue
                for pair in genshi_pairs:
                    if pair[0] == p:
                        break
                opening, closing = pair

                o_ancestors = list(opening.iterancestors())
                c_ancestors = list(closing.iterancestors())
                for n in o_ancestors:
                    if n in c_ancestors:
                        ancestor = n
                        break

                genshi_node = lxml.etree.Element('{%s}%s' % (NS['py'],
                                                             directive),
                                                 attrib={attr: a_val},
                                                 nsmap=NS)
                can_append = False
                for node in ancestor.iterchildren():
                    if node in o_ancestors:
                        outermost_o_ancestor = node
                        can_append = True
                        continue
                    if node in c_ancestors:
                        outermost_c_ancestor = node
                        break
                    if can_append:
                        genshi_node.append(node)
                ancestor.replace(outermost_o_ancestor, genshi_node)
                ancestor.remove(outermost_c_ancestor)
            else:
                p.attrib['{%s}replace' % NS['py']] = directive

    def _handle_images(self, tree):
        for draw in tree.xpath('//draw:frame', namespaces=NS):
            d_name = draw.attrib['{%s}name' % NS['draw']]
            if d_name.startswith('image: '):
                attr_expr = "make_href(%s, '%s')" % (d_name[7:], d_name[7:])
                attributes = {}
                attributes['{%s}attrs' % NS['py']] = attr_expr
                image_node = lxml.etree.Element('{%s}image' % NS['draw'],
                                                attrib=attributes,
                                                nsmap=NS)
                draw.replace(draw[0], image_node)


    def generate(self, *args, **kwargs):
        new_oo = StringIO()
        inzip = zipfile.ZipFile(self.filepath)
        outzip = zipfile.ZipFile(new_oo, 'w')

        kwargs['make_href'] = ImageHref(outzip)
        content = str(self.content_template.generate(*args, **kwargs))

        for f in inzip.infolist():
            if f.filename == 'content.xml':
                s = StringIO()
                s.write(str(content))
                outzip.writestr('content.xml', s.getvalue())
            else:
                outzip.writestr(f, inzip.read(f.filename))
        inzip.close()
        outzip.close()

        return new_oo




###############################################################################
#
# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
# Reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

__revision__ = "$Id: __init__.py 14 2008-07-14 22:09:55Z nicoe $"
__metaclass__ = type

from genshi.template import Template as GenshiTemplate, MarkupTemplate


class NullTemplate(GenshiTemplate):

    def __init__(self, source, filepath=None, filename=None, loader=None,
                 encoding=None, lookup='strict', allow_exec=True):
        raise NotImplementedError

    def _parse(self, source, encoding):
        pass
    
    def _prepare(self, stream):
        pass

    def generate(self, *args, **kwargs):
        pass



import operator
from cStringIO import StringIO
import relatorio

class Invoice(dict):

    @property
    def total(self):
        return sum(l['amount'] for l in self['lines'])

    @property
    def vat(self):
        return self.total * 0.21


repos = relatorio.ReportRepository()
repos.add_report(Invoice, 'application/vnd.oasis.opendocument.text',
                 'basic.odt', report_name='basic')
repos.add_report(Invoice, 'application/vnd.oasis.opendocument.text',
                 'invoice.odt', report_name='complicated')
repos.add_report(Invoice, 'application/vnd.oasis.opendocument.text',
                 'test.odt', report_name='test')

inv = Invoice(customer={'name': 'John Bonham',
                        'address': {'street': 'Smirnov street',
                                    'zip': 1000,
                                    'city': 'Montreux'}},
              lines=[{'item': {'name': 'Vodka 70cl',
                               'reference': 'VDKA-001',
                               'price': 10.34},
                      'quantity': 7,
                      'amount': 7*10.34},
                     {'item': {'name': 'Cognac 70cl',
                               'reference': 'CGNC-067',
                               'price': 13.46},
                      'quantity': 12,
                      'amount': 12*13.46},
                     {'item': {'name': 'Sparkling water 25cl',
                               'reference': 'WATR-007',
                               'price': 0.4},
                      'quantity': 1,
                      'amount': 0.4},
                     {'item': {'name': 'Good customer',
                               'reference': 'BONM-001',
                               'price': -20},
                      'quantity': 1,
                      'amount': -20},
                    ],
              id='MZY-20080703',
              status='late',
              trombine=(file('bouteille.png', 'r'), 'image/png'))


_, basic_report = repos.reports[Invoice]['basic']
file('bonham_basic.odt', 'w').write(basic_report(inv).getvalue())
_, report = repos.reports[Invoice]['complicated']
file('bonham_complicated.odt', 'w').write(report(inv).getvalue())



# -*- encoding: utf-8 -*-
import relatorio
from setuptools import setup, find_packages

setup(
    name="relatorio",
    url="http://relatorio.openhex.org",
    author="Nicolas Evrard",
    author_email="nicoe@openhex.org",
    description="A templating library able to output odt and pdf files",
    long_description="""
relatorio
========

A templating library which provides a way to easily output odt and pdf file.

relatorio also provides a report repository allowing you to link python objects
and report together, find reports by mimetypes/name/pyhon objects.
""",
    license="GPL License",
    version=relatorio.__version__,
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=[
        "Genshi >= 0.5"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing",
    ],
    test_suite="nose.collector")



###############################################################################
#
# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
# Reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

__revision__ = "$Id: test_odt.py 1 2008-07-04 14:31:52Z nicoe $"

import os
from cStringIO import StringIO

import lxml.etree
from nose.tools import *
from genshi.template import MarkupTemplate

from mimetypes.odt import OOTemplate, NS


class TestOOTemplating(object):

    def setup(self):
        thisdir = os.path.dirname(__file__)
        filepath = os.path.join(thisdir, 'test.odt')
        self.oot = OOTemplate(file(filepath), filepath)

    def test_init(self):
        assert_true(isinstance(self.oot.content_template, MarkupTemplate))

    def test_directives(self):
        xml = '''<a xmlns="urn:a" xmlns:text="%s">
        <text:placeholder>&lt;foo&gt;</text:placeholder>
        </a>''' % NS['text']
        parsed = self.oot.add_directives(xml) 
        root = lxml.etree.parse(StringIO(xml)).getroot()
        root_parsed = lxml.etree.parse(StringIO(parsed)).getroot()
        eq_(root_parsed[0].attrib['{http://genshi.edgewall.org/}replace'], 
            'foo')
